@@ -11,7 +11,6 @@
 #define LDD_ARGV0    "__LDD_ARGV0"
 #define DIR_SEP      ":, \t\n"
 #define MAX_DIRS     32
-#define DEFAULT_PATH "/usr/lib:/lib"
 
 typedef void (*loadptr)(int func, ...);
 typedef void (*callbackptr)(int ver, int nlibs, char **libs, 
@@ -29,22 +28,17 @@ typedef void (*callbackptr)(int ver, int nlibs, char **libs,
 
 #define LDSO_CACHE_MAGIC "ld.so-"
 #define LDSO_CACHE_MAGIC_LEN (sizeof LDSO_CACHE_MAGIC -1)
-#define LDSO_CACHE_VER "1.6.0"
+#define LDSO_CACHE_VER "1.7.0"
 #define LDSO_CACHE_VER_LEN (sizeof LDSO_CACHE_VER -1)
 
 typedef struct {
 	char magic   [LDSO_CACHE_MAGIC_LEN];
 	char version [LDSO_CACHE_VER_LEN];
-	char ndirs;
-	char nlibs;
-	long liboffset;
-	long diroffset;
+	int nlibs;
 } header_t;
 
 typedef struct {
-	char flags;
-	short libnamelen;
-	short dirnamelen;
-	long liboffset;
-	long diroffset;
+	int flags;
+	int sooffset;
+	int liboffset;
 } libentry_t;
