@@ -189,7 +189,7 @@ void _dl_boot(int args){
   struct dyn_elf * rpnt;
   struct elf_resolve * app_tpnt;
   unsigned int brk_addr;
-  unsigned int dl_data[10];
+  unsigned int dl_data[AT_EGID+1];
   unsigned char * malloc_buffer, *mmap_zero;
   int (*_dl_atexit)(void *);
   int * lpnt;
@@ -210,11 +210,12 @@ void _dl_boot(int args){
   envp = (char **) aux_dat;
   while(*aux_dat) aux_dat++;  /* Skip over the envp pointers */
   aux_dat++;  /* Skip over NULL at end of envp */
+  dl_data[AT_EUID] = -1; /* check later to see if it is changed */
   while(*aux_dat)
     {
       unsigned int * ad1;
       ad1 = aux_dat + 1;
-      if( *aux_dat <= AT_ENTRY ) dl_data[*aux_dat] = *ad1;
+      if( *aux_dat <= AT_EGID ) dl_data[*aux_dat] = *ad1;
       aux_dat += 2;
     }
 
@@ -230,20 +231,15 @@ void _dl_boot(int args){
      yet, so we cannot use real malloc. */
 
   {
+    /* This hack is to work around a suspected asm bug in gcc-2.7.0 */
     int zfileno;
-    char  * zfile = "/dev/zero";
+#define ZFILENO ((-1 & (~zfileno)) | zfileno)
+/*#define ZFILENO -1*/
+
     /* See if we need to relocate this address */
-    if( zfile < (char *) load_addr ) zfile += load_addr;
-    zfileno = _dl_open(zfile, 0);
-    if (zfileno < 0)
-    {
-	SEND_STDERR("dl_boot: open /dev/zero failed!\n");
-	_dl_exit(12);
-    }
     mmap_zero = malloc_buffer = (unsigned char *) _dl_mmap((void*) 0, 4096,
 			     PROT_READ | PROT_WRITE, 
-			     MAP_PRIVATE, zfileno, 0);
-    _dl_close(zfileno);
+			     MAP_PRIVATE | MAP_ANONYMOUS, ZFILENO, 0);
     if(_dl_mmap_check_error(mmap_zero)) {
 	SEND_STDERR("dl_boot: mmap of /dev/zero failed!\n");
 	_dl_exit(13);
@@ -497,26 +493,36 @@ void _dl_boot(int args){
   {
     _dl_not_lazy = _dl_getenv("LD_BIND_NOW",envp);
 
+    if ( (dl_data[AT_UID] == -1 && _dl_suid_ok()) ||
+	 (dl_data[AT_UID] != -1 && dl_data[AT_UID] == dl_data[AT_EUID] &&
+	  dl_data[AT_GID] == dl_data[AT_EGID]))
+    {
 #ifndef IBCS_COMPATIBLE
-    _dl_preload = _dl_getenv("LD_ELF_PRELOAD", envp);
-    if (!_dl_preload)
+      if (!(_dl_preload = _dl_getenv("LD_ELF_PRELOAD", envp)))
 #endif
-      _dl_preload = _dl_getenv("LD_PRELOAD", envp);
+	_dl_preload = _dl_getenv("LD_PRELOAD", envp);
 #ifndef IBCS_COMPATIBLE
-    _dl_library_path = _dl_getenv("LD_ELF_LIBRARY_PATH", envp);
-    if (!_dl_library_path)
+      if (!(_dl_library_path = _dl_getenv("LD_ELF_LIBRARY_PATH", envp)))
 #endif
-      _dl_library_path = _dl_getenv("LD_LIBRARY_PATH",envp);
-
-    if(!_dl_suid_ok() ) {
-      if (_dl_preload) {
-	*_dl_preload = '\0';
-	_dl_preload = 0;
-      }
-      if (_dl_library_path) {
-	*_dl_library_path = '\0';
-	_dl_library_path = 0;
-      }
+	_dl_library_path = _dl_getenv("LD_LIBRARY_PATH",envp);
+    }
+    else
+    {
+      char *cp;
+      if ((cp = _dl_getenv("LD_PRELOAD", envp)))
+	*cp = '\0';
+      if ((cp = _dl_getenv("LD_ELF_PRELOAD", envp)))
+	*cp = '\0';
+      if ((cp = _dl_getenv("LD_AOUT_PRELOAD", envp)))
+	*cp = '\0';
+      _dl_preload = NULL;
+      if ((cp = _dl_getenv("LD_LIBRARY_PATH", envp)))
+	*cp = '\0';
+      if ((cp = _dl_getenv("LD_ELF_LIBRARY_PATH", envp)))
+	*cp = '\0';
+      if ((cp = _dl_getenv("LD_AOUT_LIBRARY_PATH", envp)))
+	*cp = '\0';
+      _dl_library_path = NULL;
     }
   }
 
@@ -540,14 +546,14 @@ void _dl_boot(int args){
     if (_dl_preload)  {
       int i;
       char *str;
-      char c=1;
+      char c = _dl_preload[0];
       str = _dl_preload; 
       for (i = 0; c != 0; i++) {
 	if (_dl_preload[i] == ':' ||
 	    _dl_preload[i] == '\0') {
 	  c= _dl_preload[i];
 	  _dl_preload[i]=0;
-	  tpnt1 = _dl_load_shared_library(_dl_symbol_tables, str);
+	  tpnt1 = _dl_load_shared_library(NULL, str);
 	  if (!tpnt1) {
 	    if (_dl_trace_loaded_objects)
 	      _dl_fdprintf(1, "\t%s => not found\n", str);
@@ -600,7 +606,7 @@ void _dl_boot(int args){
 		tpnt = NULL;
 		continue;
 	      }
-	      if(!(tpnt1 = _dl_load_shared_library(_dl_symbol_tables, lpnt))) {
+	      if(!(tpnt1 = _dl_load_shared_library(tcurr, lpnt))) {
 		if (_dl_trace_loaded_objects)
 		  _dl_fdprintf(1, "\t%s => not found\n", lpnt);
 		else
@@ -732,6 +738,17 @@ void _dl_boot(int args){
 
   _dl_atexit = (int (*)(void *)) _dl_find_hash("atexit", NULL, 1, NULL, 0);
 
+  /*
+   * OK, fix one more thing - set up the debug_addr structure to point
+   * to our chain.  Later we may need to fill in more fields, but this
+   * should be enough for now.
+   */
+  debug_addr->r_map = (struct link_map *) _dl_loaded_modules;
+  debug_addr->r_version = 1;
+  debug_addr->r_ldbase = load_addr;
+  debug_addr->r_brk = (unsigned long) &_dl_debug_state;
+  _dl_debug_addr = debug_addr;
+
   for(tpnt = _dl_loaded_modules; tpnt; tpnt = tpnt->next)
     {
       /* Apparently crt1 for the application is responsible for handling this.
@@ -767,17 +784,6 @@ void _dl_boot(int args){
 #undef DL_DEBUG
    }
 
-  /*
-   * OK, fix one more thing - set up the debug_addr structure to point
-   * to our chain.  Later we may need to fill in more fields, but this
-   * should be enough for now. 
-   */
-  debug_addr->r_map = (struct link_map *) _dl_loaded_modules;
-  debug_addr->r_version = 1;
-  debug_addr->r_ldbase = load_addr;
-  debug_addr->r_brk = (unsigned long) &_dl_debug_state;
-  _dl_debug_addr = debug_addr;
-
   /* OK we are done here.  Turn out the lights, and lock up. */
   _dl_elf_main = (int (*)(int, char**, char**)) dl_data[AT_ENTRY];
 
@@ -840,18 +846,9 @@ void * _dl_malloc(int size) {
   	return (*_dl_malloc_function)(size);
 
   if(_dl_malloc_addr-_dl_mmap_zero+size>4096) {
-  	int zfileno;
-  	char  * zfile = "/dev/zero";
-  	zfileno = _dl_open(zfile, 0);
-	if (zfileno < 0)
-	{
-	  _dl_fdprintf(2, "%s: can't open '/dev/zero'\n", _dl_progname);
-	  _dl_exit(19);
-	}
 	_dl_mmap_zero = _dl_malloc_addr = (unsigned char *) _dl_mmap((void*) 0, size,
   				PROT_READ | PROT_WRITE, 
-  				MAP_PRIVATE, zfileno, 0);
-  	_dl_close(zfileno);
+  				MAP_PRIVATE | MAP_ANONYMOUS, -1, 0);
   	if(_dl_mmap_check_error(_dl_mmap_zero)) {
   	    _dl_fdprintf(2, "%s: can't map '/dev/zero'\n", _dl_progname);
   	    _dl_exit(20);
