@@ -9,6 +9,7 @@
 #include <link.h>
 #include <stdlib.h>
 #include <sys/mman.h>
+#include "sysdep.h"
 #include "hash.h"
 #include "string.h"
 #include "linuxelf.h"
@@ -32,35 +33,52 @@ static const char * dl_error_names[] = {
 	"File not found",
 	"Unable to open /dev/zero",
 	"Not an ELF file",
-	"Not i386 binary",
+#if defined (__i386__)
+ 	"Not i386 binary",
+#elif defined (__sparc__)
+	"Not sparc binary",
+#elif defined (__mc68000__)
+	"Not m68k binary",
+#else
+	"Unrecognized binary type",
+#endif
 	"Not an ELF shared library",
 	"Unable to mmap file",
 	"No dynamic section",
+#ifdef ELF_USES_RELOCA
+	"Unable to process REL relocs",
+#else
 	"Unable to process RELA relocs",
+#endif
 	"Bad handle",
 	"Unable to resolve symbol"
 };
 	
 static void dl_cleanup()
 {
-	struct dyn_elf * rpnt, *rpnt1;
+	struct dyn_elf*		d;
 	
-	for(rpnt = handles; rpnt; rpnt = rpnt1)
-	{
-		rpnt1 = rpnt->next_handle;
-		dlclose(rpnt);
-	}
+	for (d = handles; d; d = d->next_handle)
+		if (d->dyn->libtype == loaded_file && d->dyn->dynamic_info[DT_FINI])
+		{
+			(*((int(*)(void))(d->dyn->loadaddr + d->dyn->dynamic_info[DT_FINI])))();
+			d->dyn->dynamic_info[DT_FINI] = 0;
+		}
 }
 
 void * _dlopen(char * libname, int flag)
 {
-	struct elf_resolve * tpnt;
+	struct elf_resolve * tpnt, *tfrom;
 	struct dyn_elf * rpnt;
 	struct dyn_elf * dyn_chain;
+	struct dyn_elf * dpnt;
 	static int dl_init = 0;
+	char *  from;
 	void (*dl_brk)(void);
 	int (*dl_elf_init)(void);
 
+	from = __builtin_return_address(0);
+
 	/* Have the dynamic linker use the regular malloc function now */
 	if (!dl_init) {
 		dl_init++;
@@ -71,15 +89,38 @@ void * _dlopen(char * libname, int flag)
 	/* Cover the trivial case first */
 	if (!libname) return _dl_symbol_tables;
 
-	if(!(tpnt = _dl_load_shared_library(NULL, libname)))
-		return NULL;
+#ifdef USE_CACHE
+	_dl_map_cache();
+#endif
+
+	/*
+	 * Try and locate the module we were called from - we
+	 * need this so that we get the correct RPATH.
+	 */
+	tfrom = NULL;
+	for(dpnt = _dl_symbol_tables; dpnt; dpnt = dpnt->next)
+	  {
+	    tpnt = dpnt->dyn;
+	    if(   tpnt->loadaddr < from 
+	       && (   tfrom == NULL
+		   || tfrom->loadaddr < tpnt->loadaddr))
+	      tfrom = tpnt;
+	  }
+
+	if(!(tpnt = _dl_load_shared_library(tfrom, libname)))
+	{
+#ifdef USE_CACHE
+	  _dl_unmap_cache();
+#endif
+	  return NULL;
+	}
 
 	tpnt->usage_count++;
 	dyn_chain = rpnt = 
 	   (struct dyn_elf *) malloc(sizeof(struct dyn_elf));
 	_dl_memset (rpnt, 0, sizeof(*rpnt));
 	rpnt->dyn = tpnt;
-	tpnt->symbol_scope = dyn_chain;
+	if (!tpnt->symbol_scope) tpnt->symbol_scope = dyn_chain;
 	
 	rpnt->next_handle = handles;
 	handles = rpnt;
@@ -103,7 +144,7 @@ void * _dlopen(char * libname, int flag)
 		    {
 		      lpnt = tcurr->loadaddr + tcurr->dynamic_info[DT_STRTAB] + 
 			dpnt->d_un.d_val;
-		      if(!(tpnt1 = _dl_load_shared_library(_dl_symbol_tables, lpnt)))
+		      if(!(tpnt1 = _dl_load_shared_library(tcurr, lpnt)))
 			goto oops;
 
 		      rpnt->next = 
@@ -157,9 +198,17 @@ void * _dlopen(char * libname, int flag)
 		(*dl_elf_init)();
 	      }
    	}
+
+#ifdef USE_CACHE
+	_dl_unmap_cache();
+#endif
 	return (void *) dyn_chain;
+
 oops:
 	/* Something went wrong.  Clean up and return NULL. */
+#ifdef USE_CACHE
+	_dl_unmap_cache();
+#endif
 	do_dlclose (dyn_chain, 0);
 	return NULL;
 }
@@ -170,11 +219,26 @@ static int do_fixup(struct elf_resolve * tpnt, int flag)
   if(tpnt->next) goof += do_fixup(tpnt->next, flag);
 
   if(tpnt->dynamic_info[DT_REL]) {
+#ifdef ELF_USES_RELOCA
+    goof++;
+#else
     if (tpnt->init_flag & RELOCS_DONE) return goof;
     tpnt->init_flag |= RELOCS_DONE;
    
     goof += _dl_parse_relocation_information(tpnt, tpnt->dynamic_info[DT_REL],
 					     tpnt->dynamic_info[DT_RELSZ], 0);
+#endif
+  }
+  if(tpnt->dynamic_info[DT_RELA]) {
+#ifdef ELF_USES_RELOCA
+    if (tpnt->init_flag & RELOCS_DONE) return goof;
+    tpnt->init_flag |= RELOCS_DONE;
+   
+    goof += _dl_parse_relocation_information(tpnt, tpnt->dynamic_info[DT_RELA],
+					     tpnt->dynamic_info[DT_RELASZ], 0);
+#else
+    goof++;
+#endif
   }
   if(tpnt->dynamic_info[DT_JMPREL])
     {
@@ -189,10 +253,6 @@ static int do_fixup(struct elf_resolve * tpnt, int flag)
 						  tpnt->dynamic_info[DT_JMPREL],
 						  tpnt->dynamic_info[DT_PLTRELSZ], 0);
     };
-  if(tpnt->dynamic_info[DT_RELA]) {
-	/* FIXME - return some error code here */
-	goof++;
-  };
   return goof;
 }
 
@@ -207,15 +267,16 @@ void * _dlsym(void * vhandle, char * name)
 
 	handle = (struct dyn_elf *) vhandle;
 
-	for(rpnt = handle; rpnt; rpnt = rpnt->next_handle)
-		if(rpnt == handle) break;
-	
-	if (!rpnt) {
-		_dl_error_number = DL_BAD_HANDLE;
-		return NULL;
+	if (vhandle != RTLD_NEXT) {
+		for(rpnt = handle; rpnt; rpnt = rpnt->next_handle)
+			if(rpnt == handle) break;
+		if (!rpnt) {
+			_dl_error_number = DL_BAD_HANDLE;
+			return NULL;
+		}
 	}
 		
-	ret = _dl_find_hash(name, handle, 1, NULL, 0);
+	ret = _dl_find_hash(name, handle, 1, NULL, 1);
 	if (!ret)
 	{
 		_dl_error_number = DL_NO_SYMBOL;
@@ -386,7 +447,7 @@ static int foobar()
 	exit(1);
 }
 
-static int foobar1 = 0; /* Use as pointer */
+static int foobar1 = (int)foobar; /* Use as pointer */
 
 #ifndef IBCS_COMPATIBLE
 #if 1
@@ -396,6 +457,10 @@ static int foobar1 = 0; /* Use as pointer */
 #pragma weak _dl_debug_addr = foobar1
 #pragma weak _dl_error_number = foobar1
 #pragma weak _dl_load_shared_library = foobar
+#ifdef USE_CACHE
+#pragma weak _dl_map_cache = foobar
+#pragma weak _dl_unmap_cache = foobar
+#endif
 #pragma weak _dl_malloc_function = foobar1
 #pragma weak _dl_parse_relocation_information = foobar
 #pragma weak _dl_parse_lazy_relocation_information = foobar
@@ -406,6 +471,10 @@ __asm__(".weak _dl_loaded_modules; _dl_loaded_modules = foobar1");
 __asm__(".weak _dl_debug_addr; _dl_debug_addr = foobar1");
 __asm__(".weak _dl_error_number; _dl_error_number = foobar1");
 __asm__(".weak _dl_load_shared_library; _dl_load_shared_library = foobar");
+#ifdef USE_CACHE
+__asm__(".weak _dl_map_cache; _dl_map_cache = foobar");
+__asm__(".weak _dl_unmap_cache; _dl_unmap_cache = foobar");
+#endif
 __asm__(".weak _dl_malloc_function; _dl_malloc_function = foobar1");
 __asm__(".weak _dl_parse_relocation_information; _dl_parse_relocation_information = foobar");
 __asm__(".weak _dl_parse_lazy_relocation_information; _dl_parse_lazy_relocation_information = foobar");
@@ -424,22 +493,26 @@ void _dlinfo()
 	fprintf(stderr, "List of loaded modules\n");
 	/* First start with a complete list of all of the loaded files. */
 	for (tpnt = _dl_loaded_modules; tpnt; tpnt = tpnt->next)
-		fprintf(stderr,"\t%8.8x %8.8x %8.8x %s %d %s\n", tpnt->loadaddr, tpnt,
-			tpnt->symbol_scope,
+		fprintf(stderr, "\t%8.8x %8.8x %8.8x %s %d %s\n",
+			(unsigned)tpnt->loadaddr, (unsigned)tpnt,
+			(unsigned)tpnt->symbol_scope,
 			type[tpnt->libtype],
 			tpnt->usage_count, 
 			tpnt->libname);
 
 	/* Next dump the module list for the application itself */
-	fprintf(stderr,"\nModules for application (%x):\n", _dl_symbol_tables);
+	fprintf(stderr, "\nModules for application (%x):\n",
+		(unsigned)_dl_symbol_tables);
 	for (rpnt = _dl_symbol_tables; rpnt; rpnt = rpnt->next)
-		fprintf(stderr,"\t%8.8x %s\n", rpnt->dyn, rpnt->dyn->libname);
+		fprintf(stderr, "\t%8.8x %s\n", (unsigned)rpnt->dyn,
+			rpnt->dyn->libname);
 
 	for (hpnt = handles; hpnt; hpnt = hpnt->next_handle)
 	{
-		fprintf(stderr,"Modules for handle %x\n", hpnt);
+		fprintf(stderr, "Modules for handle %x\n", (unsigned)hpnt);
 		for(rpnt = hpnt; rpnt; rpnt = rpnt->next)
-			fprintf(stderr,"\t%8.8x %s\n", rpnt->dyn, rpnt->dyn->libname);
+			fprintf(stderr, "\t%8.8x %s\n", (unsigned)rpnt->dyn,
+				rpnt->dyn->libname);
 	}
 }
 
