@@ -5,15 +5,16 @@ umask 022
 echo Expect some cache has wrong version warnings if upgrading
 echo to version 1.7.x for the first time.
 
-source Version.mk
+. Version.mk
+
 PREFIX=
 
-#if [ -e /etc/ld.so.cache ] ; then
+#if [ -f /etc/ld.so.cache ] ; then
 #	echo Deleting old /etc/ld.so.cache
 #	rm -f /etc/ld.so.cache
 #fi
 
-if [ ! -e /etc/ld.so.conf ] ; then
+if [ ! -f /etc/ld.so.conf ] ; then
 	echo Creating new /etc/ld.so.conf
 	for dir in /usr/local/lib /usr/X11R6/lib /usr/X386/lib /usr/openwin/lib /lib/elf ; do
 		if [ -d $dir ] ; then
@@ -22,28 +23,36 @@ if [ ! -e /etc/ld.so.conf ] ; then
 	done
 fi
 
-echo Installing ld.so
-install -c ld-so/ld.so $PREFIX/lib/ld.so.$VERSION
-mv -f $PREFIX/lib/ld.so.$VERSION $PREFIX/lib/ld.so
-ln -f $PREFIX/lib/ld.so $PREFIX/lib/ld.so.$VERSION
+if [ -f ld-so/ld.so ] ; then
+	echo Installing ld.so
+	install -c ld-so/ld.so $PREFIX/lib/ld.so.$VERSION
+	mv -f $PREFIX/lib/ld.so.$VERSION $PREFIX/lib/ld.so
+	ln -f $PREFIX/lib/ld.so $PREFIX/lib/ld.so.$VERSION
 
-echo Installing libldso.a
-if [ -d $PREFIX/usr/i486-linuxaout/lib ] ; then
-	install -c -m 644 ld-so/libldso.a $PREFIX/usr/i486-linuxaout/lib
+	echo Installing libldso.a
+	if [ -d $PREFIX/usr/i486-linuxaout/lib ] ; then
+		install -c -m 644 ld-so/libldso.a $PREFIX/usr/i486-linuxaout/lib
+	else
+		install -c -m 644 ld-so/libldso.a $PREFIX/usr/lib
+	fi
 else
-	install -c -m 644 ld-so/libldso.a $PREFIX/usr/lib
+	echo Not installing a.out support
 fi
 
-echo installing ld-linux.so
-install -c d-link/ld-linux.so $PREFIX/lib/ld-linux.so.$VERSION
-mv -f $PREFIX/lib/ld-linux.so.$VERSION $PREFIX/lib/ld-linux.so.$VMAJOR
-ln -f $PREFIX/lib/ld-linux.so.$VMAJOR $PREFIX/lib/ld-linux.so.$VERSION
-if [ -d $PREFIX/lib/elf ] ; then
-	ln -sf ../ld-linux.so.$VMAJOR $PREFIX/lib/elf/ld-linux.so.$VMAJOR
-fi
+if [ -f d-link/ld-linux.so ] ; then
+	echo Installing ld-linux.so
+	install -c d-link/ld-linux.so $PREFIX/lib/ld-linux.so.$VERSION
+	mv -f $PREFIX/lib/ld-linux.so.$VERSION $PREFIX/lib/ld-linux.so.$VMAJOR
+	ln -f $PREFIX/lib/ld-linux.so.$VMAJOR $PREFIX/lib/ld-linux.so.$VERSION
+	if [ -d $PREFIX/lib/elf ] ; then
+		ln -sf ../ld-linux.so.$VMAJOR $PREFIX/lib/elf/ld-linux.so.$VMAJOR
+	fi
 
-echo installing libdl.so
-install -c d-link/libdl/libdl.so $PREFIX/lib/libdl.so.$VERSION
+	echo Installing libdl.so
+	install -c d-link/libdl/libdl.so $PREFIX/lib/libdl.so.$VERSION
+else
+	echo Not installing ELF support
+fi
 
 echo Installing ldd
 install -c util/ldd $PREFIX/usr/bin
