@@ -59,7 +59,7 @@
 #include <fcntl.h>
 #include <errno.h>
 #include <limits.h>
-#include <asm/string.h>
+#include <linux/string.h>
 #include <sys/stat.h>
 #include "../config.h"
 #include "fixups.h"
@@ -170,7 +170,7 @@ shared_loader(int func, ...)
     int nlibs=0;
     int nowarn;
     int keepdir;
-    char *preload;
+    char *preload = NULL;
     caddr_t cache_addr;
     size_t cache_size;
     callbackptr callback = (callbackptr)0;
@@ -229,34 +229,36 @@ shared_loader(int func, ...)
     if (!argv0 && !(argv0 = getenv(LDD_ARGV0)))
 	argv0 = LDSO_IMAGE;
 
-    /* hmm, you want your own path, do you? */
-    if (((cp = getenv("LD_AOUT_LIBRARY_PATH")) && *cp) ||
-        ((cp = getenv("LD_LIBRARY_PATH")) && *cp))
+    /* hmm, you want your own configuration, do you? */
+    if (getuid() == geteuid() && getgid() == getegid())
     {
-	uid_t uid = getuid();
-	if (uid && (uid != geteuid() || getgid() != getegid()))
-	{
-	    /* sorry, Charlie, I can't let you do that */
-	    *cp = '\0';
-	}
-	else
-	{
-	    /* Well OK, I'll allow it - this time */
-	    char *temp = alloca(strlen(cp) + 1);
-	    ndirs = add_dirs(argv0, ndirs, libdir, strcpy(temp, cp));
-	}
+      /* Well OK, I'll allow it, but just this time */
+      if (!(preload = getenv("LD_AOUT_PRELOAD")))
+        preload = getenv("LD_PRELOAD");
+      if (!(cp = getenv("LD_AOUT_LIBRARY_PATH")))
+        cp = getenv("LD_LIBRARY_PATH");
+      if (cp && *cp)
+      {
+	char *temp = alloca(strlen(cp) + 1);
+	ndirs = add_dirs(argv0, ndirs, libdir, strcpy(temp, cp));
+      }
     }
-
-    /* so you want to override something, eh? */
-    if (((preload = getenv("LD_AOUT_PRELOAD")) && *preload) ||
-        ((preload = getenv("LD_PRELOAD")) && *preload))
+    else
     {
-	uid_t uid = getuid();
-	if (uid && (uid != geteuid() || getgid() != getegid()))
-	{
-	    /* no way, Jose, that's not allowed here */
-	    *preload = '\0';
-	}
+      /* sorry, Charlie, I can't let you do that */
+      if ((cp = getenv("LD_PRELOAD")))
+	*cp = '\0';
+      if ((cp = getenv("LD_ELF_PRELOAD")))
+	*cp = '\0';
+      if ((cp = getenv("LD_AOUT_PRELOAD")))
+	*cp = '\0';
+      preload = NULL;
+      if ((cp = getenv("LD_LIBRARY_PATH")))
+	*cp = '\0';
+      if ((cp = getenv("LD_ELF_LIBRARY_PATH")))
+	*cp = '\0';
+      if ((cp = getenv("LD_AOUT_LIBRARY_PATH")))
+	*cp = '\0';
     }
 
     /* attempt to read and initialize the cache */
@@ -453,8 +455,8 @@ static caddr_t cache_init(char *file, size_t *size)
     if (memcmp(((header_t *)cache_addr)->version, LDSO_CACHE_VER, 
 	       LDSO_CACHE_VER_LEN))
     {
-	fdprintf(2, "%s: cache '%s' has wrong version\n",
-		 LDSO_IMAGE, LDSO_CACHE);
+	fdprintf(2, "%s: cache '%s' is not version %s\n",
+		 LDSO_IMAGE, LDSO_CACHE, LDSO_CACHE_VER);
 	goto no_cache;
     }
 
