@@ -14,7 +14,7 @@
  *        dir ...: directories to process
  *        lib ...: libraries to link
  *
- * Copyright 1994 David Engel and Mitch D'Souza
+ * Copyright 1994, 1995 David Engel and Mitch D'Souza
  *
  * This program may be used for any purpose as long as this
  * copyright notice is kept.
@@ -23,7 +23,7 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <stdarg.h>
-#include <asm/string.h>
+#include <string.h>
 #include <ctype.h>
 #include <getopt.h>
 #include <dirent.h>
@@ -60,19 +60,14 @@ int libmode = 0;		/* library mode */
 int nocache = 0;		/* don't build cache */
 int nolinks = 0;		/* don't update links */
 
-#define LEN_SORT	0x1
-#define NAME_SORT	0x2
-#define CTIME_SORT	0x3
-
-char sort_type = NAME_SORT;	/* default to sort by name */
 char *cachefile = LDSO_CACHE;	/* default cache file */
-signed char invert = 1;		/* Invert any of the sort options */
-
 void cache_print(void);
-void cache_dodir(char *, short);
-void cache_dolib(char *, short, char);
+void cache_dolib(char *dir, char *so, char *lib, int);
+void cache_rmlib(char *dir, char *so);
 void cache_write(void);
 
+char *readsoname(FILE *file);
+
 void warn(char *fmt, ...)
 {
     va_list ap;
@@ -120,23 +115,24 @@ char *xstrdup(char *str)
     return ptr;
 }
 
-/* if shared library, return length of base name, else return -1 */
-int is_shlib(char *dir, char *name, char *type)
+/* if shared library, return a malloced copy of the soname and set the
+   type, else return NULL */
+char *is_shlib(char *dir, char *name, int *type)
 {
-    int good = 0;
-    char *cp = name, *cp2;
+    char *good = NULL;
+    char *cp, *cp2;
     FILE *file;
     struct exec exec;
     struct elfhdr *elf_hdr;
     struct stat statbuf;
     char buff[1024];
 
-    /* see if name is of the form libZ.so.M[.N[.P]] */
+    /* see if name is of the form libZ.so* */
     if (strncmp(name, "lib", 3) == 0 &&
-	(cp = strstr(name, ".so.")) && cp[4])
+	(cp = strstr(name, ".so")))
     {
-	/* find the start of the M.P part, if any */
-	if ((cp2 = strchr(cp + 4, '.')))
+	/* find the start of the Vminor part, if any */
+	if (cp[3] == '.' && (cp2 = strchr(cp + 4, '.')))
 	    cp = cp2;
 	else
 	    cp = cp + strlen(cp);
@@ -148,15 +144,13 @@ int is_shlib(char *dir, char *name, char *type)
 	/* first, make sure it's a regular file */
 	if (lstat(buff, &statbuf))
 	    warn("can't lstat %s (%s), skipping", buff, strerror(errno));
-	else if (S_ISLNK(statbuf.st_mode))
+	else if (S_ISLNK(statbuf.st_mode) && stat(buff, &statbuf))
 	{
-	    if (stat(buff, &statbuf))
-	    {
-		warn("%s is a stale symlink, removing", buff);
-		if (remove(buff))
-		    warn("can't remove %s (%s), skipping", buff,
-			 strerror(errno));
-	    }
+	    if (verbose)
+	        warn("%s is a stale symlink, removing", buff);
+	    if (unlink(buff))
+	        warn("can't unlink %s (%s), skipping", buff,
+		     strerror(errno));
 	}
 	else if (!S_ISREG(statbuf.st_mode))
 	    warn("%s is not a regular file, skipping", buff, strerror(errno));
@@ -167,8 +161,7 @@ int is_shlib(char *dir, char *name, char *type)
 		warn("can't open %s (%s), skipping", buff, strerror(errno));
 	    else
 	    {
-		/* now make sure it's [ZQ]MAGIC with a non-zero address
-                   or ELF */
+		/* now make sure it's a DLL or ELF library */
 		if (fread(&exec, sizeof exec, 1, file) < 1)
 		    warn("can't read header from %s, skipping", buff);
 		else if (N_MAGIC(exec) != ZMAGIC && N_MAGIC(exec) != QMAGIC)
@@ -176,18 +169,32 @@ int is_shlib(char *dir, char *name, char *type)
 		    elf_hdr = (struct elfhdr *) &exec;
 		    if (elf_hdr->e_ident[0] != 0x7f ||
 			strncmp(&elf_hdr->e_ident[1], "ELF",3) != 0)
-			warn("%s is not a [ZQ]MAGIC or ELF file, skipping", buff);
+			warn("%s is not a DLL or ELF library, skipping", buff);
 		    else
 		    {
-			good = 1;
+			if ((good = readsoname(file)) == NULL)
+			    good = xstrdup(name);
+			else
+			{
+			    /* if the soname does not match the filename,
+			       we assume it is for compile-time compatibility
+			       and ignore it. */
+			    int len = strlen(good);
+			    if (strncmp(good, name, len) != 0 ||
+				(name[len] != '\0' && name[len] != '.'))
+			    {
+			        free(good);
+				good = NULL;
+			    }
+			}
 			*type = LIB_ELF;
 		    }
 		}
-		else if (!exec.a_entry)
-		    warn("%s has a zero load address, skipping", buff);
 		else
 		{
-		    good = 1;	/* looks good */
+		    good = xmalloc(cp - name + 1);
+		    strncpy(good, name, cp - name);
+		    good[cp - name] = '\0';
 		    *type = LIB_DLL;
 		}
 		fclose(file);
@@ -195,11 +202,11 @@ int is_shlib(char *dir, char *name, char *type)
 	}
     }
 
-    return good ? cp - name : -1;
+    return good;
 }
 
 /* update the symlink to new library */
-void link_shlib(char *dir, char *file, int len)
+void link_shlib(char *dir, char *file, char *so)
 {
     int change = -1;
     char libname[1024];
@@ -208,10 +215,8 @@ void link_shlib(char *dir, char *file, int len)
     struct stat linkstat;
 
     /* construct the full path names */
-    sprintf(libname, "%s%s%s", dir, (*dir && strcmp(dir, "/")) ?
-	    "/" : "", file);
-    sprintf(linkname, "%s%s%.*s", dir, (*dir && strcmp(dir, "/"))
-	    ? "/" : "", len, file);
+    sprintf(libname, "%s/%s", dir, file);
+    sprintf(linkname, "%s/%s", dir, so);
 
     /* see if a link already exists */
     if (!stat(linkname, &linkstat))
@@ -237,7 +242,7 @@ void link_shlib(char *dir, char *file, int len)
 
     /* some people like to know what we're doing */
     if (verbose)
-	printf("\t%.*s => %s%s\n", len, file, file,
+	printf("\t%s => %s%s\n", so, file,
 	       change < 0 ? " (SKIPPED)" :
 	       (change > 0 ? " (changed)" : ""));
 
@@ -269,8 +274,8 @@ int libcmp(char *p1, char *p2)
 
 struct lib
 {
+    char *so;			/* soname of a library */
     char *name;			/* name of a library */
-    int len;			/* length of base name */
     struct lib *next;		/* next library in list */
 };
 
@@ -279,17 +284,14 @@ void scan_dir(char *name)
 {
     DIR *dir;
     struct dirent *ent;
-    int len;
+    char *so;
     struct lib *lp, *libs = NULL;
-    char libtype;
+    int libtype;
 
     /* let 'em know what's going on */
     if (verbose)
 	printf("%s:\n", name);
 
-    if (!nocache)
-	cache_dodir(name, strlen(name));
-
     /* if we can't open it, we can't do anything */
     if ((dir = opendir(name)) == NULL)
     {
@@ -301,20 +303,19 @@ void scan_dir(char *name)
     while ((ent = readdir(dir)) != NULL)
     {
 	/* if it's not a shared library, don't bother */
-	if ((len = is_shlib(name, ent->d_name, &libtype)) < 0)
+	if ((so = is_shlib(name, ent->d_name, &libtype)) == NULL)
 	    continue;
 
 	if (!nocache)
-	    cache_dolib(ent->d_name, len, libtype);
+	    cache_dolib(name, so, ent->d_name, libtype);
 
-	/* have we already seen one with the same base name? */
+	/* have we already seen one with the same so name? */
 	for (lp = libs; lp; lp = lp->next)
 	{
-	    if (len == lp->len && 
-		strncmp(ent->d_name, lp->name, len) == 0)
+	    if (strcmp(so, lp->so) == 0)
 	    {
-		/* we have, which one do we want to use? */
-		if (libcmp(&ent->d_name[len+1], &lp->name[len+1]) > 0)
+	        /* we have, which one do we want to use? */
+	        if (libcmp(ent->d_name, lp->name) > 0)
 		{
 		    /* let's use the new one */
 		    free(lp->name);
@@ -328,11 +329,13 @@ void scan_dir(char *name)
 	if (!lp)
 	{
 	    lp = xmalloc(sizeof *lp);
+	    lp->so = xstrdup(so);
 	    lp->name = xstrdup(ent->d_name);
-	    lp->len = len;
 	    lp->next = libs;
 	    libs = lp;
 	}
+
+	free(so);
     }
 
     /* don't need this any more */
@@ -340,12 +343,17 @@ void scan_dir(char *name)
 
     /* now we have all the latest libs, update the links */
     for (lp = libs; lp; lp = lp->next)
-	link_shlib(name, lp->name, lp->len);
+    {
+	link_shlib(name, lp->name, lp->so);
+	if (strcmp(lp->name, lp->so) != 0)
+	    cache_rmlib(name, lp->so);
+    }
 
     /* always try to clean up after ourselves */
     while (libs)
     {
 	lp = libs->next;
+	free(libs->so);
 	free(libs->name);
 	free(libs);
 	libs = lp;
@@ -375,17 +383,17 @@ char *get_extpath(void)
 
 int main(int argc, char **argv)
 {
-    int i, c, len;
+    int i, c;
     int nodefault = 0;
     int printcache = 0;
-    char *cp, *dir;
-    char *extpath, libtype;
-    static char defpath[] = DEFAULT_PATH;
+    char *cp, *dir, *so;
+    char *extpath;
+    int libtype;
 
     prog = argv[0];
     opterr = 0;
 
-    while ((c = getopt(argc, argv, "DvnNXlpf:S:")) != EOF)
+    while ((c = getopt(argc, argv, "DvnNXlpf:")) != EOF)
 	switch (c)
 	{
 	case 'D':
@@ -425,28 +433,6 @@ int main(int argc, char **argv)
 	case 'f':
 	    cachefile = optarg;	/* alternate cache file */
 	    break;
-	case 'S':
-	    for (cp = optarg; *cp; cp++)
-	    {
-		switch (*cp)
-		{
-		case 'C':
-		    sort_type = CTIME_SORT; /* sort by lib creation time */
-		    break;
-		case 'L':
-		    sort_type = LEN_SORT; /* sort by lib length */
-		    break;
-		case 'N':
-		    sort_type = NAME_SORT; /* sort by lib name */
-		    break;
-		case 'i':
-		    invert = -1; /* invert sort order */
-		    break;
-		default:
-		    error("invalid sort option -S%s", optarg);
-		    break;
-		}
-	    }
 	}
 
     /* allow me to introduce myself, hi, my name is ... */
@@ -482,12 +468,12 @@ int main(int argc, char **argv)
 	    }
 
 	    /* we'd better do a little bit of checking */
-	    if ((len = is_shlib(dir, cp, &libtype)) < 0)
+	    if ((so = is_shlib(dir, cp, &libtype)) == NULL)
 		error("%s%s%s is not a shared library", dir,
 		      (*dir && strcmp(dir, "/")) ? "/" : "", cp);
 
 	    /* so far, so good, maybe he knows what he's doing */
-	    link_shlib(dir, cp, len);
+	    link_shlib(dir, cp, so);
 	}
     }
     else
@@ -518,9 +504,8 @@ int main(int argc, char **argv)
 	    }
 
 	    /* everybody needs these, don't they? */
-	    for (cp = strtok(defpath, DIR_SEP); cp;
-		 cp = strtok(NULL, DIR_SEP))
-		scan_dir(cp);
+	    scan_dir("/usr/lib");
+	    scan_dir("/lib");
 	}
 
 	if (!nocache)
@@ -530,171 +515,137 @@ int main(int argc, char **argv)
     exit(EXIT_OK);
 }
 
-typedef struct dirlist
-{
-    short dirnamelen;
-    long offset;
-    char *entry;
-    struct dirlist *q_forw;
-    struct dirlist *q_back;
-} dirlist_t;
-
 typedef struct liblist
 {
-    char flags;
-    short libnamelen;
-    short dirnamelen;
-    long liboffset;
-    long diroffset;
-    char *entry;
-    struct stat st;
-    struct liblist *q_forw;
-    struct liblist *q_back;
+    int flags;
+    int sooffset;
+    int liboffset;
+    char *soname;
+    char *libname;
+    int skip;
+    struct liblist *next;
 } liblist_t;
 
-#define LINK_MASK 0xff
+static header_t magic = { LDSO_CACHE_MAGIC, LDSO_CACHE_VER, 0 };
+static liblist_t *lib_head = NULL;
 
-static header_t magic;
-static dirlist_t *cur_dir = NULL;
-static liblist_t *cur_lib = NULL;
-
-void cache_dodir(char *entry, short len)
+static int liblistcomp(liblist_t *x, liblist_t *y)
 {
-    magic.ndirs++;
-    if (cur_dir == NULL)
+    int res;
+
+    if ((res = libcmp(x->soname, y->soname)) == 0)
     {
-	cur_dir = xmalloc(sizeof (dirlist_t));
-	cur_dir->dirnamelen = len;
-	cur_dir->q_forw = NULL;
-	cur_dir->q_back = NULL;
-	cur_dir->offset = 0;
-	cur_dir->entry = xstrdup(entry);
+        res = libcmp(strrchr(x->libname, '/') + 1,
+		     strrchr(y->libname, '/') + 1);
+    }
+
+    return res;
+}
+
+void cache_dolib(char *dir, char *so, char *lib, int libtype)
+{
+    char fullpath[PATH_MAX];
+    liblist_t *new_lib, *cur_lib;
+
+    magic.nlibs++;
+    sprintf(fullpath, "%s/%s", dir, lib);
+    new_lib = xmalloc(sizeof (liblist_t));
+    new_lib->flags = libtype;
+    new_lib->soname = xstrdup(so);
+    new_lib->libname = xstrdup(fullpath);
+    new_lib->skip = 0;
+
+    if (lib_head == NULL || liblistcomp(new_lib, lib_head) > 0)
+    {
+        new_lib->next = lib_head;
+	lib_head = new_lib;
     }
     else
     {
-	cur_dir->q_forw = xmalloc(sizeof (dirlist_t));
-	cur_dir->q_forw->dirnamelen = len;
-	cur_dir->q_forw->q_forw = NULL;
-	cur_dir->q_forw->q_back = cur_dir;
-	cur_dir->q_forw->offset = cur_dir->offset
-	    + cur_dir->dirnamelen + 1;
-	cur_dir->q_forw->entry = xstrdup(entry);
-	cur_dir = cur_dir->q_forw;
+        for (cur_lib = lib_head; cur_lib->next != NULL &&
+	     liblistcomp(new_lib, cur_lib->next) <= 0;
+	     cur_lib = cur_lib->next)
+	    /* nothing */;
+	new_lib->next = cur_lib->next;
+	cur_lib->next = new_lib;
     }
 }
 
-void cache_dolib(char *entry, short len, char libtype)
+void cache_rmlib(char *dir, char *so)
 {
     char fullpath[PATH_MAX];
+    liblist_t *cur_lib;
 
-    magic.nlibs++;
-    if (cur_lib == NULL)
+    sprintf(fullpath, "%s/%s", dir, so);
+
+    for (cur_lib = lib_head; cur_lib != NULL; cur_lib = cur_lib->next)
     {
-	sprintf(fullpath, "%s/%s", cur_dir->entry, entry);
-	cur_lib = (liblist_t *) xmalloc(sizeof (liblist_t));
-	cur_lib->flags = libtype;
-	cur_lib->libnamelen = strlen(entry)+(len<<8);
-	cur_lib->dirnamelen = cur_dir->dirnamelen;
-	cur_lib->diroffset = cur_dir->offset;
-	cur_lib->liboffset = 0;
-	stat(fullpath,&cur_lib->st);
-	cur_lib->q_forw = NULL;
-	cur_lib->q_back = NULL;
-	cur_lib->entry = xstrdup(entry);
-    }
-    else
-    {
-	sprintf(fullpath, "%s/%s", cur_dir->entry, entry);
-	cur_lib->q_forw = (liblist_t *) xmalloc(sizeof (liblist_t));
-	cur_lib->q_forw->flags = libtype;
-	cur_lib->q_forw->libnamelen = strlen(entry)+(len<<8);
-	cur_lib->q_forw->dirnamelen = cur_dir->dirnamelen;
-	cur_lib->q_forw->diroffset = cur_dir->offset;
-	cur_lib->q_forw->liboffset = cur_lib->liboffset +
-	    (cur_lib->libnamelen & LINK_MASK) + 1;
-	stat(fullpath,&cur_lib->q_forw->st);
-	cur_lib->q_forw->q_forw = NULL;
-	cur_lib->q_forw->q_back = cur_lib;
-	cur_lib->q_forw->entry = xstrdup(entry);
-	cur_lib = cur_lib->q_forw;
+        if (strcmp(cur_lib->soname, so) == 0 &&
+	    strcmp(cur_lib->libname, fullpath) == 0 &&
+	    !cur_lib->skip)
+	{
+	    magic.nlibs--;
+	    cur_lib->skip = 1;
+	}
     }
 }
 
 void cache_write(void)
 {
-    int i, cachefd;
+    int cachefd;
+    int stroffset = 0;
     char tempfile[1024];
-    void sort_cache(liblist_t *, int);
+    liblist_t *cur_lib;
 
     if (!magic.nlibs)
 	return;
 
     sprintf(tempfile, "%s~", cachefile);
 
-    if ((cachefd = open(tempfile, O_RDWR|O_CREAT|O_TRUNC, 0666)) < 0)
-	error("can't open %s (%s)", tempfile, strerror(errno));
+    if (unlink(tempfile) && errno != ENOENT)
+        error("can't unlink %s (%s)", tempfile, strerror(errno));
 
-    lseek(cachefd, sizeof (header_t), SEEK_SET);
+    if ((cachefd = creat(tempfile, 0644)) < 0)
+	error("can't create %s (%s)", tempfile, strerror(errno));
 
-    sort_cache(cur_lib, magic.nlibs);
+    if (write(cachefd, &magic, sizeof (header_t)) != sizeof (header_t))
+	error("can't write %s (%s)", tempfile, strerror(errno));
 
-    while (cur_lib->q_back)
-	cur_lib = cur_lib->q_back;
-
-    for (i = 0; i < magic.nlibs; i++)
+    for (cur_lib = lib_head; cur_lib != NULL; cur_lib = cur_lib->next)
     {
-	write(cachefd, cur_lib, sizeof (libentry_t));
-	if (cur_lib->q_forw == NULL)
-	    break;
-	else
-	    cur_lib = cur_lib->q_forw;
+        if (!cur_lib->skip)
+	{
+	    cur_lib->sooffset = stroffset;
+	    stroffset += strlen(cur_lib->soname) + 1;
+	    cur_lib->liboffset = stroffset;
+	    stroffset += strlen(cur_lib->libname) + 1;
+	    if (write(cachefd, cur_lib, sizeof (libentry_t)) !=
+		sizeof (libentry_t))
+	        error("can't write %s (%s)", tempfile, strerror(errno));
+	}
     }
 
-    while (cur_lib->q_back)
-	cur_lib = cur_lib->q_back;
-
-    for (i = 0; i < magic.nlibs; i++)
+    for (cur_lib = lib_head; cur_lib != NULL; cur_lib = cur_lib->next)
     {
-	write(cachefd, cur_lib->entry, (cur_lib->libnamelen & LINK_MASK) + 1);
-
-	if (cur_lib->q_forw == NULL)
-	    break;
-	else
-	    cur_lib = cur_lib->q_forw;
-
-	free (cur_lib->q_back->entry);
-	free (cur_lib->q_back);
+        if (!cur_lib->skip)
+	{
+	    if (write(cachefd, cur_lib->soname, strlen(cur_lib->soname) + 1)
+		!= strlen(cur_lib->soname) + 1)
+	        error("can't write %s (%s)", tempfile, strerror(errno));
+	    if (write(cachefd, cur_lib->libname, strlen(cur_lib->libname) + 1)
+		!= strlen(cur_lib->libname) + 1)
+	        error("can't write %s (%s)", tempfile, strerror(errno));
+	}
     }
 
-    magic.diroffset = lseek(cachefd, 0L, SEEK_CUR);
+    if (close(cachefd))
+        error("can't close %s (%s)", tempfile, strerror(errno));
 
-    while (cur_dir->q_back)
-	cur_dir = cur_dir->q_back;
-
-    while (cur_dir)
-    {
-	write(cachefd, cur_dir->entry, cur_dir->dirnamelen + 1);
-
-	if (cur_dir->q_forw == NULL)
-	    break;
-	else
-	    cur_dir = cur_dir->q_forw;
-
-	free (cur_dir->q_back->entry);
-	free (cur_dir->q_back);
-    }
-
-    lseek(cachefd, 0L, SEEK_SET);
-    memcpy(&magic, LDSO_CACHE_MAGIC LDSO_CACHE_VER,
-	   sizeof magic.magic + sizeof magic.version);
-    magic.liboffset = sizeof (header_t)
-	+ magic.nlibs * sizeof (libentry_t);
-    write(cachefd, &magic, sizeof (header_t));
-
-    close(cachefd);
+    if (chmod(tempfile, 0644))
+	error("can't chmod %s (%s)", tempfile, strerror(errno));
 
     if (rename(tempfile, cachefile))
-	error("can't rename %s to %s (%s)", tempfile, cachefile, strerror(errno));
+	error("can't rename %s (%s)", tempfile, strerror(errno));
 }
 
 void cache_print(void)
@@ -702,9 +653,9 @@ void cache_print(void)
     caddr_t c;
     struct stat st;
     int fd = 0;
-    char nlibs;
-    char *liboffset, *diroffset;
-    libentry_t libent;
+    char *strs;
+    header_t *header;
+    libentry_t *libent;
 
     if (stat(cachefile, &st) || (fd = open(cachefile, O_RDONLY))<0)
 	error("can't read %s (%s)", cachefile, strerror(errno));
@@ -718,103 +669,21 @@ void cache_print(void)
     if (memcmp(((header_t *)c)->version, LDSO_CACHE_VER, LDSO_CACHE_VER_LEN))
 	error("wrong cache version - expected %s", LDSO_CACHE_VER);
 
-    nlibs = ((header_t *)c)->nlibs;
-    liboffset = c+((header_t *)c)->liboffset;
-    diroffset = c+((header_t *)c)->diroffset;
+    header = (header_t *)c;
+    libent = (libentry_t *)(c + sizeof (header_t));
+    strs = (char *)&libent[header->nlibs];
 
-    fprintf(stdout,"%d libs found in cache `%s' (version %s)\n", nlibs, cachefile, LDSO_CACHE_VER);
+    printf("%d libs found in cache `%s' (version %s)\n",
+	    header->nlibs, cachefile, LDSO_CACHE_VER);
 
-    for (fd = 0; fd<nlibs; fd++)
+    for (fd = 0; fd < header->nlibs; fd++)
     {
-	memcpy(&libent, c+sizeof(header_t)+fd * sizeof(libentry_t),
-	       sizeof(libentry_t));
-	fprintf(stdout, "\t%2d - ",fd+1);
-	fflush(stdout);
-	write(1,diroffset+libent.diroffset,libent.dirnamelen);
-	write(1,"/",1);
-	write(1,liboffset+libent.liboffset,libent.libnamelen & LINK_MASK);
-	write(1, libent.flags == LIB_DLL ? " (DLL)":" (ELF)",6);
-	write(1,"\n",1);
+	printf("\t%2d - %s %s => %s\n", fd + 1, 
+	libent[fd].flags == LIB_DLL ? "DLL" : "ELF",
+	strs + libent[fd].sooffset, 
+	strs + libent[fd].liboffset);
     }
+
     munmap (c,st.st_size);
 }
 
-static int libnamecomp(const void *x, const void *y)
-{
-#if 1
-    return invert * libcmp((*(liblist_t **)y)->entry,
-			   (*(liblist_t **)x)->entry);
-#else
-    return invert * strncmp((*(liblist_t **)y)->entry,
-			    (*(liblist_t **)x)->entry,
-			    (*(liblist_t **)y)->libnamelen & LINK_MASK);
-#endif
-}
-
-static int liblencomp(const void *x, const void *y)
-{
-    return invert * (((*(liblist_t **)y)->libnamelen & LINK_MASK) -
-		     ((*(liblist_t **)x)->libnamelen & LINK_MASK));
-}
-
-static int libctimecomp(const void *x, const void *y)
-{
-    return invert * (((*(liblist_t **)y)->st.st_ctime) -
-		     ((*(liblist_t **)x)->st.st_ctime));
-}
-
-void sort_cache(liblist_t *cur_lib, int nlibs)
-{
-    int i;
-    char **array;
-
-    while (cur_lib->q_back)
-	cur_lib = cur_lib->q_back;
-
-    array = (char **)malloc(nlibs * sizeof(char *));
-
-    for (i = 0; i < nlibs; i++)
-    {
-	array[i] = (char *)cur_lib;
-	if (!cur_lib->q_forw) break;
-	cur_lib = cur_lib->q_forw;
-    }
-
-    switch (sort_type)
-    {
-    case LEN_SORT:
-	qsort(array, nlibs, sizeof (char *), liblencomp);
-	break;
-    case NAME_SORT:
-	qsort(array, nlibs, sizeof (char *), libnamecomp);
-	break;
-    case CTIME_SORT:
-	qsort(array, nlibs, sizeof (char *), libctimecomp);
-	break;
-    }
-
-    while (cur_lib->q_back)
-	cur_lib = cur_lib->q_back;
-    
-    /* Relink the sorted linked list and calculate the new offsets */
-    for (i = 0; i < nlibs; i++)
-	if (i == 0)
-	{
-	    ((liblist_t *)array[i])->q_back = NULL;
-	    ((liblist_t *)array[i])->q_forw = NULL;
-	    ((liblist_t *)array[i])->liboffset = 0;
-	}
-	else
-	{
-	    ((liblist_t *)array[i-1])->q_forw = ((liblist_t *)array[i]);
-	    ((liblist_t *)array[i])->q_back = ((liblist_t *)array[i-1]);
-	    ((liblist_t *)array[i])->q_forw = NULL;
-	    ((liblist_t *)array[i])->liboffset = 
-		(long)((liblist_t *)array[i])->q_back->liboffset +
-		    ((long)((liblist_t *)array[i])->q_back->libnamelen & LINK_MASK) + 1;
-	}
-
-    cur_lib = (liblist_t *)array;
-
-    free(array);
-}
