@@ -17,7 +17,7 @@
 
 #include <stdio.h>
 #include <stdlib.h>
-#include <asm/string.h>
+#include <string.h>
 #include <getopt.h>
 #include <unistd.h>
 #include <a.out.h>
